@@ -14,6 +14,7 @@ use crate::{
         borrow::DormantMutRef,
         fmt_utils::StrDisplayAsDebug,
         item_set::ItemSet,
+        map_hash::MapHash,
     },
 };
 use alloc::{collections::BTreeSet, vec::Vec};
@@ -80,7 +81,7 @@ pub struct BiHashMap<T, S = DefaultHashBuilder, A: Allocator = Global> {
     pub(super) items: ItemSet<T, A>,
     // Invariant: the values (usize) in these tables are valid indexes into
     // `items`, and are a 1:1 mapping.
-    tables: BiHashMapTables<S, A>,
+    pub(super) tables: BiHashMapTables<S, A>,
 }
 
 impl<T: BiHashItem, S: Default, A: Allocator + Default> Default
@@ -179,7 +180,7 @@ impl<T: BiHashItem> BiHashMap<T> {
     }
 }
 
-impl<T: BiHashItem, S: Clone + BuildHasher> BiHashMap<T, S> {
+impl<T: BiHashItem, S: BuildHasher> BiHashMap<T, S> {
     /// Creates a new `BiHashMap` with the given hasher.
     ///
     /// # Examples
@@ -212,14 +213,10 @@ impl<T: BiHashItem, S: Clone + BuildHasher> BiHashMap<T, S> {
     /// let map: BiHashMap<Item, RandomState> = BiHashMap::with_hasher(hasher);
     /// assert!(map.is_empty());
     /// ```
-    pub fn with_hasher(hasher: S) -> Self {
+    pub const fn with_hasher(hasher: S) -> Self {
         Self {
             items: ItemSet::new(),
-            tables: BiHashMapTables::with_capacity_and_hasher_in(
-                0,
-                hasher,
-                global_alloc(),
-            ),
+            tables: BiHashMapTables::with_hasher(hasher),
         }
     }
 
@@ -668,6 +665,52 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         self.items.len()
     }
 
+    /// Clears the map, removing all items.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[cfg(feature = "default-hasher")] {
+    /// use iddqd::{BiHashItem, BiHashMap, bi_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq)]
+    /// struct Item {
+    ///     id: u32,
+    ///     name: String,
+    ///     value: i32,
+    /// }
+    ///
+    /// impl BiHashItem for Item {
+    ///     type K1<'a> = u32;
+    ///     type K2<'a> = &'a str;
+    ///
+    ///     fn key1(&self) -> Self::K1<'_> {
+    ///         self.id
+    ///     }
+    ///     fn key2(&self) -> Self::K2<'_> {
+    ///         &self.name
+    ///     }
+    ///     bi_upcast!();
+    /// }
+    ///
+    /// let mut map = BiHashMap::new();
+    /// map.insert_unique(Item { id: 1, name: "foo".to_string(), value: 42 })
+    ///     .unwrap();
+    /// map.insert_unique(Item { id: 2, name: "bar".to_string(), value: 99 })
+    ///     .unwrap();
+    /// assert_eq!(map.len(), 2);
+    ///
+    /// map.clear();
+    /// assert!(map.is_empty());
+    /// assert_eq!(map.len(), 0);
+    /// # }
+    /// ```
+    pub fn clear(&mut self) {
+        self.items.clear();
+        self.tables.k1_to_item.clear();
+        self.tables.k2_to_item.clear();
+    }
+
     /// Returns an iterator over all items in the map.
     ///
     /// Similar to [`HashMap`], the iteration order is arbitrary and not
@@ -1069,9 +1112,10 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state.clone();
         let hashes =
             awakened_map.tables.make_hashes::<T>(&item.key1(), &item.key2());
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     /// Removes the item uniquely identified by `key1` and `key2`, if it exists.
@@ -1205,9 +1249,10 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state.clone();
         let hashes =
             awakened_map.tables.make_hashes::<T>(&item.key1(), &item.key2());
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     /// Removes an item from the map by its `key1`.
@@ -1411,9 +1456,10 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state.clone();
         let hashes =
             awakened_map.tables.make_hashes::<T>(&item.key1(), &item.key2());
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     /// Removes an item from the map by its `key2`.
@@ -1552,14 +1598,16 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         let (index1, index2) = {
             // index1 and index2 are explicitly typed to show that it has a
             // trivial Drop impl that doesn't capture anything from map.
-            let index1: Option<usize> = map
-                .tables
-                .k1_to_item
-                .find_index(&key1, |index| map.items[index].key1());
-            let index2: Option<usize> = map
-                .tables
-                .k2_to_item
-                .find_index(&key2, |index| map.items[index].key2());
+            let index1: Option<usize> = map.tables.k1_to_item.find_index(
+                &map.tables.state,
+                &key1,
+                |index| map.items[index].key1(),
+            );
+            let index2: Option<usize> = map.tables.k2_to_item.find_index(
+                &map.tables.state,
+                &key2,
+                |index| map.items[index].key2(),
+            );
             (index1, index2)
         };
 
@@ -1596,6 +1644,123 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         }
     }
 
+    /// Retains only the elements specified by the predicate.
+    ///
+    /// In other words, remove all items `T` for which `f(RefMut<T>)` returns
+    /// false. The elements are visited in an arbitrary order.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[cfg(feature = "default-hasher")] {
+    /// use iddqd::{BiHashItem, BiHashMap, bi_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, Hash)]
+    /// struct Item {
+    ///     id: u32,
+    ///     name: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl BiHashItem for Item {
+    ///     type K1<'a> = u32;
+    ///     type K2<'a> = &'a str;
+    ///
+    ///     fn key1(&self) -> Self::K1<'_> {
+    ///         self.id
+    ///     }
+    ///     fn key2(&self) -> Self::K2<'_> {
+    ///         &self.name
+    ///     }
+    ///
+    ///     bi_upcast!();
+    /// }
+    ///
+    /// let mut map = BiHashMap::new();
+    /// map.insert_unique(Item { id: 1, name: "foo".to_string(), value: 42 })
+    ///     .unwrap();
+    /// map.insert_unique(Item { id: 2, name: "bar".to_string(), value: 20 })
+    ///     .unwrap();
+    /// map.insert_unique(Item { id: 3, name: "baz".to_string(), value: 99 })
+    ///     .unwrap();
+    ///
+    /// // Retain only items where value is greater than 30
+    /// map.retain(|item| item.value > 30);
+    ///
+    /// assert_eq!(map.len(), 2);
+    /// assert_eq!(map.get1(&1).unwrap().value, 42);
+    /// assert_eq!(map.get1(&3).unwrap().value, 99);
+    /// assert!(map.get1(&2).is_none());
+    /// # }
+    /// ```
+    pub fn retain<'a, F>(&'a mut self, mut f: F)
+    where
+        F: FnMut(RefMut<'a, T, S>) -> bool,
+    {
+        let hash_state = self.tables.state.clone();
+        let (_, mut dormant_items) = DormantMutRef::new(&mut self.items);
+
+        self.tables.k1_to_item.retain(|index| {
+            let (item, dormant_items) = {
+                // SAFETY: All uses of `items` ended in the previous iteration.
+                let items = unsafe { dormant_items.reborrow() };
+                let (items, dormant_items) = DormantMutRef::new(items);
+                let item: &'a mut T = items
+                    .get_mut(index)
+                    .expect("all indexes are present in self.items");
+                (item, dormant_items)
+            };
+
+            let (hashes, dormant_item) = {
+                let (item, dormant_item): (&'a mut T, _) =
+                    DormantMutRef::new(item);
+                // Use T::k1(item) rather than item.key() to force the key
+                // trait function to be called for T rather than &mut T.
+                let key1 = T::key1(item);
+                let key2 = T::key2(item);
+                let hash1 = hash_state.hash_one(key1);
+                let hash2 = hash_state.hash_one(key2);
+                ([MapHash::new(hash1), MapHash::new(hash2)], dormant_item)
+            };
+
+            // SAFETY: The original items is no longer used after the first
+            // block above.
+            let items = unsafe { dormant_items.awaken() };
+            // SAFETY: The original item is no longer used after the second
+            // block above.
+            let item = unsafe { dormant_item.awaken() };
+
+            let hash2 = hashes[1].hash();
+
+            let ref_mut = RefMut::new(hash_state.clone(), hashes, item);
+            if f(ref_mut) {
+                true
+            } else {
+                items.remove(index);
+                let k2_entry = self
+                    .tables
+                    .k2_to_item
+                    .find_entry_by_hash(hash2, |map2_index| {
+                        map2_index == index
+                    });
+                match k2_entry {
+                    Ok(entry) => {
+                        entry.remove();
+                    }
+                    Err(_) => {
+                        // This happening means there's an inconsistency between
+                        // the maps.
+                        panic!(
+                            "inconsistency between k1_to_item and k2_to_item"
+                        );
+                    }
+                }
+
+                false
+            }
+        });
+    }
+
     fn find1<'a, Q>(&'a self, k: &Q) -> Option<&'a T>
     where
         Q: Hash + Equivalent<T::K1<'a>> + ?Sized,
@@ -1607,7 +1772,9 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
     where
         Q: Hash + Equivalent<T::K1<'a>> + ?Sized,
     {
-        self.tables.k1_to_item.find_index(k, |index| self.items[index].key1())
+        self.tables
+            .k1_to_item
+            .find_index(&self.tables.state, k, |index| self.items[index].key1())
     }
 
     fn find2<'a, Q>(&'a self, k: &Q) -> Option<&'a T>
@@ -1621,7 +1788,9 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
     where
         Q: Hash + Equivalent<T::K2<'a>> + ?Sized,
     {
-        self.tables.k2_to_item.find_index(k, |index| self.items[index].key2())
+        self.tables
+            .k2_to_item
+            .find_index(&self.tables.state, k, |index| self.items[index].key2())
     }
 
     pub(super) fn get_by_entry_index(
@@ -1649,31 +1818,35 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         match indexes.disjoint_keys() {
             DisjointKeys::Unique(index) => {
                 let item = self.items.get_mut(index).expect("index is valid");
+                let state = self.tables.state.clone();
                 let hashes =
                     self.tables.make_hashes::<T>(&item.key1(), &item.key2());
-                OccupiedEntryMut::Unique(RefMut::new(hashes, item))
+                OccupiedEntryMut::Unique(RefMut::new(state, hashes, item))
             }
             DisjointKeys::Key1(index1) => {
                 let item =
                     self.items.get_mut(index1).expect("key1 index is valid");
+                let state = self.tables.state.clone();
                 let hashes =
                     self.tables.make_hashes::<T>(&item.key1(), &item.key2());
                 OccupiedEntryMut::NonUnique {
-                    by_key1: Some(RefMut::new(hashes, item)),
+                    by_key1: Some(RefMut::new(state, hashes, item)),
                     by_key2: None,
                 }
             }
             DisjointKeys::Key2(index2) => {
                 let item =
                     self.items.get_mut(index2).expect("key2 index is valid");
+                let state = self.tables.state.clone();
                 let hashes =
                     self.tables.make_hashes::<T>(&item.key1(), &item.key2());
                 OccupiedEntryMut::NonUnique {
                     by_key1: None,
-                    by_key2: Some(RefMut::new(hashes, item)),
+                    by_key2: Some(RefMut::new(state, hashes, item)),
                 }
             }
             DisjointKeys::Key12(indexes) => {
+                let state = self.tables.state.clone();
                 let mut items = self.items.get_disjoint_mut(indexes);
                 let item1 = items[0].take().expect("key1 index is valid");
                 let item2 = items[1].take().expect("key2 index is valid");
@@ -1683,8 +1856,8 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
                     self.tables.make_hashes::<T>(&item2.key1(), &item2.key2());
 
                 OccupiedEntryMut::NonUnique {
-                    by_key1: Some(RefMut::new(hashes1, item1)),
-                    by_key2: Some(RefMut::new(hashes2, item2)),
+                    by_key1: Some(RefMut::new(state.clone(), hashes1, item1)),
+                    by_key2: Some(RefMut::new(state, hashes2, item2)),
                 }
             }
         }
@@ -1695,10 +1868,11 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         index: usize,
     ) -> Option<RefMut<'_, T, S>> {
         let borrowed = self.items.get_mut(index)?;
+        let state = self.tables.state.clone();
         let hashes =
             self.tables.make_hashes::<T>(&borrowed.key1(), &borrowed.key2());
         let item = &mut self.items[index];
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     pub(super) fn insert_unique_impl(
@@ -1710,6 +1884,7 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         // Check for duplicates *before* inserting the new item, because we
         // don't want to partially insert the new item and then have to roll
         // back.
+        let state = &self.tables.state;
         let (e1, e2) = {
             let k1 = value.key1();
             let k2 = value.key2();
@@ -1717,13 +1892,13 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
             let e1 = detect_dup_or_insert(
                 self.tables
                     .k1_to_item
-                    .entry(k1, |index| self.items[index].key1()),
+                    .entry(state, k1, |index| self.items[index].key1()),
                 &mut duplicates,
             );
             let e2 = detect_dup_or_insert(
                 self.tables
                     .k2_to_item
-                    .entry(k2, |index| self.items[index].key2()),
+                    .entry(state, k2, |index| self.items[index].key2()),
                 &mut duplicates,
             );
             (e1, e2)
@@ -1778,8 +1953,9 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
         let value = self.items.remove(remove_index)?;
 
         // Remove the value from the tables.
+        let state = &self.tables.state;
         let Ok(item1) =
-            self.tables.k1_to_item.find_entry(&value.key1(), |index| {
+            self.tables.k1_to_item.find_entry(state, &value.key1(), |index| {
                 if index == remove_index {
                     value.key1()
                 } else {
@@ -1791,7 +1967,7 @@ impl<T: BiHashItem, S: Clone + BuildHasher, A: Allocator> BiHashMap<T, S, A> {
             panic!("remove_index {remove_index} not found in k1_to_item");
         };
         let Ok(item2) =
-            self.tables.k2_to_item.find_entry(&value.key2(), |index| {
+            self.tables.k2_to_item.find_entry(state, &value.key2(), |index| {
                 if index == remove_index {
                     value.key2()
                 } else {
