@@ -9,10 +9,14 @@ use crate::{
         alloc::{Global, global_alloc},
         borrow::DormantMutRef,
         item_set::ItemSet,
+        map_hash::MapHash,
     },
 };
 use alloc::collections::BTreeSet;
-use core::{fmt, hash::Hash};
+use core::{
+    fmt,
+    hash::{BuildHasher, Hash},
+};
 use equivalent::{Comparable, Equivalent};
 
 /// An ordered map where the keys are part of the values, based on a B-Tree.
@@ -310,6 +314,43 @@ impl<T: IdOrdItem> IdOrdMap<T> {
         self.items.len()
     }
 
+    /// Clears the map, removing all items.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "foo".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bar".to_string(), value: 99 }).unwrap();
+    /// assert_eq!(map.len(), 2);
+    ///
+    /// map.clear();
+    /// assert!(map.is_empty());
+    /// assert_eq!(map.len(), 0);
+    /// ```
+    pub fn clear(&mut self) {
+        self.items.clear();
+        self.tables.key_to_item.clear();
+    }
+
     /// Iterates over the items in the map.
     ///
     /// Similar to [`BTreeMap`], the iteration is ordered by [`T::Key`].
@@ -675,6 +716,7 @@ impl<T: IdOrdItem> IdOrdMap<T> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state().clone();
         let (hash, dormant) = {
             let (item, dormant) = DormantMutRef::new(item);
             let hash = awakened_map.tables.make_hash(item);
@@ -683,7 +725,7 @@ impl<T: IdOrdItem> IdOrdMap<T> {
 
         // SAFETY: the original item is not used after this point.
         let item = unsafe { dormant.awaken() };
-        Some(RefMut::new(hash, item))
+        Some(RefMut::new(state, hash, item))
     }
 
     /// Removes an item from the map by its `key`.
@@ -819,6 +861,365 @@ impl<T: IdOrdItem> IdOrdMap<T> {
         )
     }
 
+    /// Returns the first item in the map. The key of this item is the minimum
+    /// key in the map.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "charlie".to_string(), value: 30 }).unwrap();
+    /// map.insert_unique(Item { id: "alice".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bob".to_string(), value: 99 }).unwrap();
+    ///
+    /// // First item has the minimum key.
+    /// let first = map.first().unwrap();
+    /// assert_eq!(first.id, "alice");
+    /// assert_eq!(first.value, 42);
+    ///
+    /// // Empty map returns None.
+    /// let empty_map: IdOrdMap<Item> = IdOrdMap::new();
+    /// assert!(empty_map.first().is_none());
+    /// ```
+    #[inline]
+    pub fn first(&self) -> Option<&T> {
+        self.tables.key_to_item.first().map(|index| &self.items[index])
+    }
+
+    /// Returns the first entry in the map for in-place manipulation. The key of
+    /// this entry is the minimum key in the map.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "charlie".to_string(), value: 30 }).unwrap();
+    /// map.insert_unique(Item { id: "alice".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bob".to_string(), value: 99 }).unwrap();
+    ///
+    /// // Modify the first entry.
+    /// if let Some(mut entry) = map.first_entry() {
+    ///     entry.get_mut().value = 100;
+    /// }
+    ///
+    /// assert_eq!(map.get("alice").unwrap().value, 100);
+    /// ```
+    pub fn first_entry(&mut self) -> Option<OccupiedEntry<'_, T>> {
+        let index = self.tables.key_to_item.first()?;
+        let (_, dormant_map) = DormantMutRef::new(self);
+        Some(
+            // SAFETY: `map` is dropped immediately while creating the
+            // DormantMutRef.
+            unsafe { OccupiedEntry::new(dormant_map, index) },
+        )
+    }
+
+    /// Removes and returns the first element in the map. The key of this
+    /// element is the minimum key in the map.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "charlie".to_string(), value: 30 }).unwrap();
+    /// map.insert_unique(Item { id: "alice".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bob".to_string(), value: 99 }).unwrap();
+    ///
+    /// // Remove the first element.
+    /// let first = map.pop_first().unwrap();
+    /// assert_eq!(first.id, "alice");
+    /// assert_eq!(first.value, 42);
+    /// assert_eq!(map.len(), 2);
+    ///
+    /// // Remove the next element.
+    /// let first = map.pop_first().unwrap();
+    /// assert_eq!(first.id, "bob");
+    ///
+    /// // Empty map returns None.
+    /// map.pop_first();
+    /// assert!(map.pop_first().is_none());
+    /// ```
+    pub fn pop_first(&mut self) -> Option<T> {
+        let index = self.tables.key_to_item.first()?;
+        self.remove_by_index(index)
+    }
+
+    /// Returns the last item in the map. The key of this item is the maximum
+    /// key in the map.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "charlie".to_string(), value: 30 }).unwrap();
+    /// map.insert_unique(Item { id: "alice".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bob".to_string(), value: 99 }).unwrap();
+    ///
+    /// // Last item has the maximum key.
+    /// let last = map.last().unwrap();
+    /// assert_eq!(last.id, "charlie");
+    /// assert_eq!(last.value, 30);
+    ///
+    /// // Empty map returns None.
+    /// let empty_map: IdOrdMap<Item> = IdOrdMap::new();
+    /// assert!(empty_map.last().is_none());
+    /// ```
+    #[inline]
+    pub fn last(&self) -> Option<&T> {
+        self.tables.key_to_item.last().map(|index| &self.items[index])
+    }
+
+    /// Returns the last entry in the map for in-place manipulation. The key of
+    /// this entry is the maximum key in the map.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "charlie".to_string(), value: 30 }).unwrap();
+    /// map.insert_unique(Item { id: "alice".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bob".to_string(), value: 99 }).unwrap();
+    ///
+    /// // Modify the last entry.
+    /// if let Some(mut entry) = map.last_entry() {
+    ///     entry.get_mut().value = 200;
+    /// }
+    ///
+    /// assert_eq!(map.get("charlie").unwrap().value, 200);
+    /// ```
+    pub fn last_entry(&mut self) -> Option<OccupiedEntry<'_, T>> {
+        let index = self.tables.key_to_item.last()?;
+        let (_, dormant_map) = DormantMutRef::new(self);
+        Some(
+            // SAFETY: `map` is dropped immediately while creating the
+            // DormantMutRef.
+            unsafe { OccupiedEntry::new(dormant_map, index) },
+        )
+    }
+
+    /// Removes and returns the last element in the map. The key of this
+    /// element is the maximum key in the map.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "charlie".to_string(), value: 30 }).unwrap();
+    /// map.insert_unique(Item { id: "alice".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bob".to_string(), value: 99 }).unwrap();
+    ///
+    /// // Remove the last element.
+    /// let last = map.pop_last().unwrap();
+    /// assert_eq!(last.id, "charlie");
+    /// assert_eq!(last.value, 30);
+    /// assert_eq!(map.len(), 2);
+    ///
+    /// // Remove the next element.
+    /// let last = map.pop_last().unwrap();
+    /// assert_eq!(last.id, "bob");
+    ///
+    /// // Empty map returns None.
+    /// map.pop_last();
+    /// assert!(map.pop_last().is_none());
+    /// ```
+    pub fn pop_last(&mut self) -> Option<T> {
+        let index = self.tables.key_to_item.last()?;
+        self.remove_by_index(index)
+    }
+
+    /// Retains only the elements specified by the predicate.
+    ///
+    /// In other words, remove all items `T` for which `f(RefMut<T>)` returns
+    /// false. The elements are visited in ascending key order.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use iddqd::{IdOrdItem, IdOrdMap, id_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
+    /// struct Item {
+    ///     id: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl IdOrdItem for Item {
+    ///     type Key<'a> = &'a str;
+    ///
+    ///     fn key(&self) -> Self::Key<'_> {
+    ///         &self.id
+    ///     }
+    ///
+    ///     id_upcast!();
+    /// }
+    ///
+    /// let mut map = IdOrdMap::new();
+    /// map.insert_unique(Item { id: "foo".to_string(), value: 42 }).unwrap();
+    /// map.insert_unique(Item { id: "bar".to_string(), value: 20 }).unwrap();
+    /// map.insert_unique(Item { id: "baz".to_string(), value: 99 }).unwrap();
+    ///
+    /// // Retain only items where value is greater than 30
+    /// map.retain(|item| item.value > 30);
+    ///
+    /// assert_eq!(map.len(), 2);
+    /// assert_eq!(map.get("foo").unwrap().value, 42);
+    /// assert_eq!(map.get("baz").unwrap().value, 99);
+    /// assert!(map.get("bar").is_none());
+    /// ```
+    pub fn retain<'a, F>(&'a mut self, mut f: F)
+    where
+        F: FnMut(RefMut<'a, T>) -> bool,
+        T::Key<'a>: Hash,
+    {
+        let hash_state = self.tables.state().clone();
+        let (_, mut dormant_items) = DormantMutRef::new(&mut self.items);
+
+        self.tables.key_to_item.retain(|index| {
+            let (item, dormant_items) = {
+                // SAFETY: All uses of `items` ended in the previous iteration.
+                let items = unsafe { dormant_items.reborrow() };
+                let (items, dormant_items) = DormantMutRef::new(items);
+                let item: &'a mut T = items
+                    .get_mut(index)
+                    .expect("all indexes are present in self.items");
+                (item, dormant_items)
+            };
+
+            let (hash, dormant_item) = {
+                let (item, dormant_item): (&'a mut T, _) =
+                    DormantMutRef::new(item);
+                // Use T::key(item) rather than item.key() to force the key
+                // trait function to be called for T rather than &mut T.
+                let key = T::key(item);
+                let hash = hash_state.hash_one(key);
+                (MapHash::new(hash), dormant_item)
+            };
+
+            // SAFETY: The original items is no longer used after the first
+            // block above.
+            let items = unsafe { dormant_items.awaken() };
+            // SAFETY: The original item is no longer used after the second
+            // block above.
+            let item = unsafe { dormant_item.awaken() };
+
+            let ref_mut = RefMut::new(hash_state.clone(), hash, item);
+            if f(ref_mut) {
+                true
+            } else {
+                items.remove(index);
+                false
+            }
+        });
+    }
+
     fn find<'a, Q>(&'a self, k: &Q) -> Option<&'a T>
     where
         Q: ?Sized + Comparable<T::Key<'a>>,
@@ -853,6 +1254,7 @@ impl<T: IdOrdItem> IdOrdMap<T> {
     where
         T::Key<'a>: Hash,
     {
+        let state = self.tables.state().clone();
         let (hash, dormant) = {
             let item: &'a mut T = self.items.get_mut(index)?;
             let (item, dormant) = DormantMutRef::new(item);
@@ -862,7 +1264,7 @@ impl<T: IdOrdItem> IdOrdMap<T> {
 
         // SAFETY: item is no longer used after the above point.
         let item = unsafe { dormant.awaken() };
-        Some(RefMut::new(hash, item))
+        Some(RefMut::new(state, hash, item))
     }
 
     pub(super) fn insert_unique_impl(
