@@ -8,6 +8,7 @@ use crate::{
         borrow::DormantMutRef,
         fmt_utils::StrDisplayAsDebug,
         item_set::ItemSet,
+        map_hash::MapHash,
     },
 };
 use alloc::{collections::BTreeSet, vec::Vec};
@@ -195,7 +196,7 @@ impl<T: TriHashItem> TriHashMap<T> {
     }
 }
 
-impl<T: TriHashItem, S: Clone + BuildHasher> TriHashMap<T, S> {
+impl<T: TriHashItem, S: BuildHasher> TriHashMap<T, S> {
     /// Creates a new, empty `TriHashMap` with the given hasher.
     ///
     /// # Examples
@@ -233,14 +234,10 @@ impl<T: TriHashItem, S: Clone + BuildHasher> TriHashMap<T, S> {
     ///     TriHashMap::with_hasher(RandomState::new());
     /// assert!(map.is_empty());
     /// ```
-    pub fn with_hasher(hasher: S) -> Self {
+    pub const fn with_hasher(hasher: S) -> Self {
         Self {
             items: ItemSet::new(),
-            tables: TriHashMapTables::with_capacity_and_hasher_in(
-                0,
-                hasher.clone(),
-                global_alloc(),
-            ),
+            tables: TriHashMapTables::with_hasher(hasher),
         }
     }
 
@@ -750,6 +747,61 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
         self.items.len()
     }
 
+    /// Clears the map, removing all items.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[cfg(feature = "default-hasher")] {
+    /// use iddqd::{TriHashItem, TriHashMap, tri_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq)]
+    /// struct Person {
+    ///     id: u32,
+    ///     email: String,
+    ///     phone: String,
+    ///     name: String,
+    /// }
+    ///
+    /// impl TriHashItem for Person {
+    ///     type K1<'a> = u32;
+    ///     type K2<'a> = &'a str;
+    ///     type K3<'a> = &'a str;
+    ///
+    ///     fn key1(&self) -> Self::K1<'_> {
+    ///         self.id
+    ///     }
+    ///     fn key2(&self) -> Self::K2<'_> {
+    ///         &self.email
+    ///     }
+    ///     fn key3(&self) -> Self::K3<'_> {
+    ///         &self.phone
+    ///     }
+    ///     tri_upcast!();
+    /// }
+    ///
+    /// let mut map = TriHashMap::new();
+    /// map.insert_unique(Person {
+    ///     id: 1,
+    ///     email: "alice@example.com".to_string(),
+    ///     phone: "555-1234".to_string(),
+    ///     name: "Alice".to_string(),
+    /// })
+    /// .unwrap();
+    /// assert_eq!(map.len(), 1);
+    ///
+    /// map.clear();
+    /// assert!(map.is_empty());
+    /// assert_eq!(map.len(), 0);
+    /// # }
+    /// ```
+    pub fn clear(&mut self) {
+        self.items.clear();
+        self.tables.k1_to_item.clear();
+        self.tables.k2_to_item.clear();
+        self.tables.k3_to_item.clear();
+    }
+
     /// Iterates over the items in the map.
     ///
     /// Similar to [`HashMap`], the iteration order is arbitrary and not
@@ -1102,6 +1154,7 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
         // Check for duplicates *before* inserting the new item, because we
         // don't want to partially insert the new item and then have to roll
         // back.
+        let state = &self.tables.state;
         let (e1, e2, e3) = {
             let k1 = value.key1();
             let k2 = value.key2();
@@ -1110,19 +1163,19 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
             let e1 = detect_dup_or_insert(
                 self.tables
                     .k1_to_item
-                    .entry(k1, |index| self.items[index].key1()),
+                    .entry(state, k1, |index| self.items[index].key1()),
                 &mut duplicates,
             );
             let e2 = detect_dup_or_insert(
                 self.tables
                     .k2_to_item
-                    .entry(k2, |index| self.items[index].key2()),
+                    .entry(state, k2, |index| self.items[index].key2()),
                 &mut duplicates,
             );
             let e3 = detect_dup_or_insert(
                 self.tables
                     .k3_to_item
-                    .entry(k3, |index| self.items[index].key3()),
+                    .entry(state, k3, |index| self.items[index].key3()),
                 &mut duplicates,
             );
             (e1, e2, e3)
@@ -1366,8 +1419,9 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state.clone();
         let hashes = awakened_map.tables.make_hashes(&item);
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     /// Removes the item uniquely identified by `key1`, `key2`, and `key3`, if
@@ -1621,8 +1675,9 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state.clone();
         let hashes = awakened_map.tables.make_hashes(&item);
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     /// Removes an item from the map by its `key1`.
@@ -1857,8 +1912,9 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state.clone();
         let hashes = awakened_map.tables.make_hashes(&item);
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     /// Removes an item from the map by its `key2`.
@@ -2093,8 +2149,9 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
         // SAFETY: `map` is not used after this point.
         let awakened_map = unsafe { dormant_map.awaken() };
         let item = &mut awakened_map.items[index];
+        let state = awakened_map.tables.state.clone();
         let hashes = awakened_map.tables.make_hashes(&item);
-        Some(RefMut::new(hashes, item))
+        Some(RefMut::new(state, hashes, item))
     }
 
     /// Removes an item from the map by its `key3`.
@@ -2161,6 +2218,175 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
         awakened_map.remove_by_index(remove_index)
     }
 
+    /// Retains only the elements specified by the predicate.
+    ///
+    /// In other words, remove all items `T` for which `f(RefMut<T>)` returns
+    /// false. The elements are visited in an arbitrary order.
+    ///
+    /// The `RefMut<T, S>` wrapper allows mutable access to the item while
+    /// enforcing that the three keys (`K1`, `K2`, `K3`) remain unchanged. If
+    /// a key is modified during iteration, the method will panic.
+    ///
+    /// # Performance considerations
+    ///
+    /// This method may leave the internal storage fragmented. If you need
+    /// compact storage afterward, call `make_compact()`.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # #[cfg(feature = "default-hasher")] {
+    /// use iddqd::{TriHashItem, TriHashMap, tri_upcast};
+    ///
+    /// #[derive(Debug, PartialEq, Eq, Hash)]
+    /// struct Item {
+    ///     id: u32,
+    ///     name: String,
+    ///     code: String,
+    ///     value: u32,
+    /// }
+    ///
+    /// impl TriHashItem for Item {
+    ///     type K1<'a> = u32;
+    ///     type K2<'a> = &'a str;
+    ///     type K3<'a> = &'a str;
+    ///
+    ///     fn key1(&self) -> Self::K1<'_> {
+    ///         self.id
+    ///     }
+    ///     fn key2(&self) -> Self::K2<'_> {
+    ///         &self.name
+    ///     }
+    ///     fn key3(&self) -> Self::K3<'_> {
+    ///         &self.code
+    ///     }
+    ///
+    ///     tri_upcast!();
+    /// }
+    ///
+    /// let mut map = TriHashMap::new();
+    /// map.insert_unique(Item {
+    ///     id: 1,
+    ///     name: "foo".to_string(),
+    ///     code: "x".to_string(),
+    ///     value: 42,
+    /// })
+    /// .unwrap();
+    /// map.insert_unique(Item {
+    ///     id: 2,
+    ///     name: "bar".to_string(),
+    ///     code: "y".to_string(),
+    ///     value: 20,
+    /// })
+    /// .unwrap();
+    /// map.insert_unique(Item {
+    ///     id: 3,
+    ///     name: "baz".to_string(),
+    ///     code: "z".to_string(),
+    ///     value: 99,
+    /// })
+    /// .unwrap();
+    ///
+    /// // Retain only items where value is greater than 30
+    /// map.retain(|item| item.value > 30);
+    ///
+    /// assert_eq!(map.len(), 2);
+    /// assert_eq!(map.get1(&1).unwrap().value, 42);
+    /// assert_eq!(map.get1(&3).unwrap().value, 99);
+    /// assert!(map.get1(&2).is_none());
+    /// # }
+    /// ```
+    pub fn retain<'a, F>(&'a mut self, mut f: F)
+    where
+        F: FnMut(RefMut<'a, T, S>) -> bool,
+    {
+        let hash_state = self.tables.state.clone();
+        let (_, mut dormant_items) = DormantMutRef::new(&mut self.items);
+
+        self.tables.k1_to_item.retain(|index| {
+            let (item, dormant_items) = {
+                // SAFETY: All uses of `items` ended in the previous iteration.
+                let items = unsafe { dormant_items.reborrow() };
+                let (items, dormant_items) = DormantMutRef::new(items);
+                let item: &'a mut T = items
+                    .get_mut(index)
+                    .expect("all indexes are present in self.items");
+                (item, dormant_items)
+            };
+
+            let (hashes, dormant_item) = {
+                let (item, dormant_item): (&'a mut T, _) =
+                    DormantMutRef::new(item);
+                // Use T::k1(item) rather than item.key() to force the key
+                // trait function to be called for T rather than &mut T.
+                let key1 = T::key1(item);
+                let key2 = T::key2(item);
+                let key3 = T::key3(item);
+                let hash1 = hash_state.hash_one(key1);
+                let hash2 = hash_state.hash_one(key2);
+                let hash3 = hash_state.hash_one(key3);
+                (
+                    [
+                        MapHash::new(hash1),
+                        MapHash::new(hash2),
+                        MapHash::new(hash3),
+                    ],
+                    dormant_item,
+                )
+            };
+
+            // SAFETY: The original items is no longer used after the first
+            // block above.
+            let items = unsafe { dormant_items.awaken() };
+            // SAFETY: The original item is no longer used after the second
+            // block above.
+            let item = unsafe { dormant_item.awaken() };
+
+            let hash2 = hashes[1].hash();
+            let hash3 = hashes[2].hash();
+
+            let ref_mut = RefMut::new(hash_state.clone(), hashes, item);
+            if f(ref_mut) {
+                true
+            } else {
+                items.remove(index);
+                let k2_entry = self
+                    .tables
+                    .k2_to_item
+                    .find_entry_by_hash(hash2, |map2_index| {
+                        map2_index == index
+                    });
+                let k3_entry = self
+                    .tables
+                    .k3_to_item
+                    .find_entry_by_hash(hash3, |map3_index| {
+                        map3_index == index
+                    });
+
+                let mut is_inconsistent = false;
+                if let Ok(k2_entry) = k2_entry {
+                    k2_entry.remove();
+                } else {
+                    is_inconsistent = true;
+                }
+                if let Ok(k3_entry) = k3_entry {
+                    k3_entry.remove();
+                } else {
+                    is_inconsistent = true;
+                }
+                if is_inconsistent {
+                    // This happening means there's an inconsistency among
+                    // the maps.
+                    panic!(
+                        "inconsistency among k1_to_item, k2_to_item, k3_to_item"
+                    );
+                }
+
+                false
+            }
+        });
+    }
+
     fn find1<'a, Q>(&'a self, k: &Q) -> Option<&'a T>
     where
         Q: Hash + Equivalent<T::K1<'a>> + ?Sized,
@@ -2172,7 +2398,9 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
     where
         Q: Hash + Equivalent<T::K1<'a>> + ?Sized,
     {
-        self.tables.k1_to_item.find_index(k, |index| self.items[index].key1())
+        self.tables
+            .k1_to_item
+            .find_index(&self.tables.state, k, |index| self.items[index].key1())
     }
 
     fn find2<'a, Q>(&'a self, k: &Q) -> Option<&'a T>
@@ -2186,7 +2414,9 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
     where
         Q: Hash + Equivalent<T::K2<'a>> + ?Sized,
     {
-        self.tables.k2_to_item.find_index(k, |index| self.items[index].key2())
+        self.tables
+            .k2_to_item
+            .find_index(&self.tables.state, k, |index| self.items[index].key2())
     }
 
     fn find3<'a, Q>(&'a self, k: &Q) -> Option<&'a T>
@@ -2200,15 +2430,18 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
     where
         Q: Hash + Equivalent<T::K3<'a>> + ?Sized,
     {
-        self.tables.k3_to_item.find_index(k, |index| self.items[index].key3())
+        self.tables
+            .k3_to_item
+            .find_index(&self.tables.state, k, |index| self.items[index].key3())
     }
 
     pub(super) fn remove_by_index(&mut self, remove_index: usize) -> Option<T> {
         let value = self.items.remove(remove_index)?;
 
         // Remove the value from the tables.
+        let state = &self.tables.state;
         let Ok(item1) =
-            self.tables.k1_to_item.find_entry(&value.key1(), |index| {
+            self.tables.k1_to_item.find_entry(state, &value.key1(), |index| {
                 if index == remove_index {
                     value.key1()
                 } else {
@@ -2220,7 +2453,7 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
             panic!("remove_index {remove_index} not found in k1_to_item");
         };
         let Ok(item2) =
-            self.tables.k2_to_item.find_entry(&value.key2(), |index| {
+            self.tables.k2_to_item.find_entry(state, &value.key2(), |index| {
                 if index == remove_index {
                     value.key2()
                 } else {
@@ -2232,7 +2465,7 @@ impl<T: TriHashItem, S: Clone + BuildHasher, A: Allocator> TriHashMap<T, S, A> {
             panic!("remove_index {remove_index} not found in k2_to_item")
         };
         let Ok(item3) =
-            self.tables.k3_to_item.find_entry(&value.key3(), |index| {
+            self.tables.k3_to_item.find_entry(state, &value.key3(), |index| {
                 if index == remove_index {
                     value.key3()
                 } else {
