@@ -11,7 +11,7 @@ use iddqd_test_utils::{
     },
 };
 use proptest::prelude::*;
-use std::path::Path;
+use std::{borrow::Cow, path::Path};
 use test_strategy::{Arbitrary, proptest};
 
 #[derive(Clone, Debug)]
@@ -60,9 +60,9 @@ fn debug_impls() {
 fn debug_impls_borrowed() {
     let before = bi_hash_map! {
         HashBuilder;
-        BorrowedItem { key1: "a", key2: b"b0", key3: Path::new("path0") },
-        BorrowedItem { key1: "b", key2: b"b1", key3: Path::new("path1") },
-        BorrowedItem { key1: "c", key2: b"b2", key3: Path::new("path2") },
+        BorrowedItem { key1: "a", key2: Cow::Borrowed(b"b0"), key3: Path::new("path0") },
+        BorrowedItem { key1: "b", key2: Cow::Borrowed(b"b1"), key3: Path::new("path1") },
+        BorrowedItem { key1: "c", key2: Cow::Borrowed(b"b2"), key3: Path::new("path2") },
     };
 
     assert_eq!(
@@ -76,9 +76,9 @@ fn debug_impls_borrowed() {
 
         let after = bi_hash_map! {
             HashBuilder;
-            BorrowedItem { key1: "a", key2: b"b0", key3: Path::new("path0") },
-            BorrowedItem { key1: "c", key2: b"b3", key3: Path::new("path3") },
-            BorrowedItem { key1: "d", key2: b"b4", key3: Path::new("path4") },
+            BorrowedItem { key1: "a", key2: Cow::Borrowed(b"b0"), key3: Path::new("path0") },
+            BorrowedItem { key1: "c", key2: Cow::Borrowed(b"b3"), key3: Path::new("path3") },
+            BorrowedItem { key1: "d", key2: Cow::Borrowed(b"b4"), key3: Path::new("path4") },
         };
 
         let diff = before.diff(&after).by_unique();
@@ -196,30 +196,71 @@ fn test_insert_overwrite() {
     map.validate(ValidateCompact::NonCompact).expect("validation failed");
 }
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+enum CompactnessChange {
+    /// The operation makes the map non-compact.
+    NoLongerCompact,
+    /// The operation makes the map compact.
+    BecomesCompact,
+    /// The operation doesn't change compactness.
+    NoChange,
+}
+
+impl CompactnessChange {
+    /// Applies this compactness change to the given compactness state.
+    fn apply(self, compactness: ValidateCompact) -> ValidateCompact {
+        match (compactness, self) {
+            (ValidateCompact::Compact, CompactnessChange::NoLongerCompact) => {
+                ValidateCompact::NonCompact
+            }
+            (
+                ValidateCompact::NonCompact,
+                CompactnessChange::BecomesCompact,
+            ) => ValidateCompact::Compact,
+            _ => compactness,
+        }
+    }
+}
+
 #[derive(Debug, Arbitrary)]
 enum Operation {
     // Make inserts a bit more common to try and fill up the map.
-    #[weight(3)]
+    #[weight(4)]
     InsertUnique(TestItem),
-    #[weight(2)]
+    #[weight(3)]
     InsertOverwrite(TestItem),
+    #[weight(2)]
     Get1(u8),
+    #[weight(2)]
     Get2(char),
+    #[weight(2)]
     Remove1(u8),
+    #[weight(2)]
     Remove2(char),
+    #[weight(2)]
+    RetainValueContains(char, bool),
+    #[weight(2)]
+    RetainModulo(#[strategy(0..3_u8)] u8, #[strategy(1..4_u8)] u8, bool),
+    Clear,
 }
 
 impl Operation {
-    fn remains_compact(&self) -> bool {
+    fn compactness_change(&self) -> CompactnessChange {
         match self {
             Operation::InsertUnique(_)
             | Operation::Get1(_)
-            | Operation::Get2(_) => true,
+            | Operation::Get2(_) => CompactnessChange::NoChange,
             // The act of removing items, including calls to insert_overwrite,
             // can make the map non-compact.
             Operation::InsertOverwrite(_)
             | Operation::Remove1(_)
-            | Operation::Remove2(_) => false,
+            | Operation::Remove2(_)
+            | Operation::RetainValueContains(_, _)
+            | Operation::RetainModulo(_, _, _) => {
+                CompactnessChange::NoLongerCompact
+            }
+            // Clear always makes the map compact (empty).
+            Operation::Clear => CompactnessChange::BecomesCompact,
         }
     }
 }
@@ -237,9 +278,7 @@ fn proptest_ops(
 
     // Now perform the operations on both maps.
     for op in ops.into_iter() {
-        if compactness == ValidateCompact::Compact && !op.remains_compact() {
-            compactness = ValidateCompact::NonCompact;
-        }
+        compactness = op.compactness_change().apply(compactness);
 
         match op {
             Operation::InsertUnique(item) => {
@@ -254,7 +293,13 @@ fn proptest_ops(
                 if let Err(map_err) = map_res {
                     let naive_err = naive_res.unwrap_err();
                     assert_eq!(map_err.new_item(), naive_err.new_item());
-                    assert_eq!(map_err.duplicates(), naive_err.duplicates(),);
+                    // The duplicates may be in any order, so sort them before
+                    // comparing.
+                    let mut map_err_dups = map_err.duplicates().to_vec();
+                    let mut naive_err_dups = naive_err.duplicates().to_vec();
+                    map_err_dups.sort();
+                    naive_err_dups.sort();
+                    assert_eq!(map_err_dups, naive_err_dups);
                 }
 
                 map.validate(compactness).expect("map should be valid");
@@ -297,6 +342,35 @@ fn proptest_ops(
                 assert_eq!(map_res, naive_res);
                 map.validate(compactness).expect("map should be valid");
             }
+            Operation::RetainValueContains(ch, equals) => {
+                map.retain(|item| {
+                    let contains = item.value.contains(ch);
+                    if equals { contains } else { !contains }
+                });
+                naive_map.retain(|item| {
+                    let contains = item.value.contains(ch);
+                    if equals { contains } else { !contains }
+                });
+                map.validate(compactness).expect("map should be valid");
+            }
+            Operation::RetainModulo(a, b, equals) => {
+                let modulo = a + b;
+                let remainder = a;
+                map.retain(|item| {
+                    let matches = item.key1 % modulo == remainder;
+                    if equals { matches } else { !matches }
+                });
+                naive_map.retain(|item| {
+                    let matches = item.key1 % modulo == remainder;
+                    if equals { matches } else { !matches }
+                });
+                map.validate(compactness).expect("map should be valid");
+            }
+            Operation::Clear => {
+                map.clear();
+                naive_map.clear();
+                map.validate(compactness).expect("map should be valid");
+            }
         }
 
         // Check that the iterators work correctly.
@@ -590,10 +664,16 @@ fn insert_entry_panics_for_non_matching_keys() {
 #[test]
 fn borrowed_item() {
     let mut map = BiHashMap::<BorrowedItem, HashBuilder, Alloc>::default();
-    let item1 =
-        BorrowedItem { key1: "foo", key2: b"foo", key3: Path::new("foo") };
-    let item2 =
-        BorrowedItem { key1: "bar", key2: b"bar", key3: Path::new("bar") };
+    let item1 = BorrowedItem {
+        key1: "foo",
+        key2: Cow::Borrowed(b"foo"),
+        key3: Path::new("foo"),
+    };
+    let item2 = BorrowedItem {
+        key1: "bar",
+        key2: Cow::Borrowed(b"bar"),
+        key3: Path::new("bar"),
+    };
 
     // Insert items.
     map.insert_unique(item1.clone()).unwrap();
@@ -616,6 +696,17 @@ fn borrowed_item() {
         format!("{map:?}")
     }
 
+    #[cfg(feature = "serde")]
+    fn serialize_as_map(
+        map: &BiHashMap<BorrowedItem<'_>, HashBuilder, Alloc>,
+    ) -> Result<String, iddqd_test_utils::serde_json::Error> {
+        let mut out: Vec<u8> = Vec::new();
+        let mut ser = iddqd_test_utils::serde_json::Serializer::new(&mut out);
+        bi_hash_map::BiHashMapAsMap::serialize(map, &mut ser)?;
+        Ok(String::from_utf8(out)
+            .expect("serde_json should always emit valid UTF-8"))
+    }
+
     static DEBUG_OUTPUT: &str = "{{k1: \"foo\", k2: [102, 111, 111]}: BorrowedItem { \
         key1: \"foo\", key2: [102, 111, 111], key3: \"foo\" }, \
         {k1: \"bar\", k2: [98, 97, 114]}: BorrowedItem { \
@@ -623,6 +714,166 @@ fn borrowed_item() {
 
     assert_eq!(format!("{map:?}"), DEBUG_OUTPUT);
     assert_eq!(fmt_debug(&map), DEBUG_OUTPUT);
+
+    #[cfg(feature = "serde")]
+    {
+        let map_string = serialize_as_map(&map).unwrap();
+        let deserialized: BiHashMap<BorrowedItem<'_>, HashBuilder, Alloc> =
+            iddqd_test_utils::serde_json::from_str(&map_string).unwrap();
+        assert_eq!(map, deserialized);
+    }
+}
+
+#[test]
+fn test_retain_all() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(1, 'a', "x", "foo")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "bar")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "baz")).unwrap();
+
+    let original_len = map.len();
+    map.retain(|_| true);
+
+    assert_eq!(map.len(), original_len);
+    assert_eq!(map.len(), 3);
+    map.get1(&TestKey1::new(&1)).expect("key1=1 should be present");
+    map.get1(&TestKey1::new(&2)).expect("key1=2 should be present");
+    map.get1(&TestKey1::new(&3)).expect("key1=3 should be present");
+}
+
+#[test]
+fn test_retain_none() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(1, 'a', "x", "foo")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "bar")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "baz")).unwrap();
+
+    map.retain(|_| false);
+
+    assert_eq!(map.len(), 0);
+    assert!(map.is_empty());
+}
+
+#[test]
+fn test_retain_value_contains() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(1, 'a', "x", "foo")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "bar")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "baz")).unwrap();
+    map.insert_unique(TestItem::new(4, 'd', "w", "qux")).unwrap();
+
+    map.retain(|item| item.value.contains('a'));
+
+    assert_eq!(map.len(), 2);
+    map.get1(&TestKey1::new(&2)).expect("key1=2 (bar) should be present");
+    map.get1(&TestKey1::new(&3)).expect("key1=3 (baz) should be present");
+    assert!(
+        map.get1(&TestKey1::new(&1)).is_none(),
+        "key1=1 (foo) should be removed"
+    );
+    assert!(
+        map.get1(&TestKey1::new(&4)).is_none(),
+        "key1=4 (qux) should be removed"
+    );
+}
+
+#[test]
+fn test_retain_modulo() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(0, 'a', "x", "v0")).unwrap();
+    map.insert_unique(TestItem::new(1, 'b', "y", "v1")).unwrap();
+    map.insert_unique(TestItem::new(2, 'c', "z", "v2")).unwrap();
+    map.insert_unique(TestItem::new(3, 'd', "w", "v3")).unwrap();
+    map.insert_unique(TestItem::new(4, 'e', "u", "v4")).unwrap();
+    map.insert_unique(TestItem::new(5, 'f', "t", "v5")).unwrap();
+
+    map.retain(|item| item.key1 % 3 == 1);
+
+    assert_eq!(map.len(), 2);
+    map.get1(&TestKey1::new(&1)).expect("key1=1 should be present");
+    map.get1(&TestKey1::new(&4)).expect("key1=4 should be present");
+    assert!(map.get1(&TestKey1::new(&0)).is_none(), "key1=0 should be removed");
+    assert!(map.get1(&TestKey1::new(&2)).is_none(), "key1=2 should be removed");
+    assert!(map.get1(&TestKey1::new(&3)).is_none(), "key1=3 should be removed");
+    assert!(map.get1(&TestKey1::new(&5)).is_none(), "key1=5 should be removed");
+
+    // Test with a larger map for miri coverage.
+    let mut large_map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    for i in 0..32_u8 {
+        large_map
+            .insert_unique(TestItem::new(i, char::from(b'a' + i), "y", "z"))
+            .unwrap();
+    }
+
+    large_map.retain(|item| item.key1 % 7 == 3);
+
+    for i in 0..32_u8 {
+        if i % 7 == 3 {
+            large_map
+                .get1(&TestKey1::new(&i))
+                .unwrap_or_else(|| panic!("key1={} should be present", i));
+        } else {
+            assert!(
+                large_map.get1(&TestKey1::new(&i)).is_none(),
+                "key1={} should be removed",
+                i
+            );
+        }
+    }
+}
+
+#[test]
+fn test_retain_empty_map() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.retain(|_| true);
+    assert!(map.is_empty());
+}
+
+#[test]
+fn test_clear_empty_map() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.clear();
+    assert!(map.is_empty());
+    map.validate(ValidateCompact::Compact)
+        .expect("empty cleared map should be compact");
+}
+
+#[test]
+fn test_clear_makes_compact() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+
+    // Add items
+    map.insert_unique(TestItem::new(1, 'a', "x", "v1")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "v2")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "v3")).unwrap();
+
+    // Remove an item to make it non-compact
+    map.remove1(&TestKey1::new(&2));
+    map.validate(ValidateCompact::NonCompact)
+        .expect("map should be valid but non-compact");
+
+    // Clear should make it compact again
+    map.clear();
+    assert!(map.is_empty());
+    map.validate(ValidateCompact::Compact)
+        .expect("cleared map should be compact");
+}
+
+#[test]
+fn test_retain_verifies_both_keys() {
+    let mut map = BiHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(1, 'a', "x", "foo")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "bar")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "baz")).unwrap();
+
+    // Retain only key1=2
+    map.retain(|item| item.key1 == 2);
+
+    // Verify both key1 and key2 work
+    map.get1(&TestKey1::new(&2)).expect("key1=2 should be present");
+    map.get2(&TestKey2::new('b')).expect("key2='b' should be present");
+    assert!(map.get1(&TestKey1::new(&1)).is_none());
+    assert!(map.get2(&TestKey2::new('a')).is_none());
 }
 
 mod macro_tests {
