@@ -11,7 +11,7 @@ use iddqd_test_utils::{
     },
 };
 use proptest::prelude::*;
-use std::path::Path;
+use std::{borrow::Cow, path::Path};
 use test_strategy::{Arbitrary, proptest};
 
 #[derive(Clone, Debug)]
@@ -52,9 +52,9 @@ fn debug_impls() {
 fn debug_impls_borrowed() {
     let before = id_hash_map! {
         HashBuilder;
-        BorrowedItem { key1: "a", key2: b"b0", key3: Path::new("path0") },
-        BorrowedItem { key1: "b", key2: b"b1", key3: Path::new("path1") },
-        BorrowedItem { key1: "c", key2: b"b2", key3: Path::new("path2") },
+        BorrowedItem { key1: "a", key2: Cow::Borrowed(b"b0"), key3: Path::new("path0") },
+        BorrowedItem { key1: "b", key2: Cow::Borrowed(b"b1"), key3: Path::new("path1") },
+        BorrowedItem { key1: "c", key2: Cow::Borrowed(b"b2"), key3: Path::new("path2") },
     };
 
     assert_eq!(
@@ -68,9 +68,9 @@ fn debug_impls_borrowed() {
 
         let after = id_hash_map! {
             HashBuilder;
-            BorrowedItem { key1: "a", key2: b"b0", key3: Path::new("path0") },
-            BorrowedItem { key1: "c", key2: b"b3", key3: Path::new("path3") },
-            BorrowedItem { key1: "d", key2: b"b4", key3: Path::new("path4") },
+            BorrowedItem { key1: "a", key2: Cow::Borrowed(b"b0"), key3: Path::new("path0") },
+            BorrowedItem { key1: "c", key2: Cow::Borrowed(b"b3"), key3: Path::new("path3") },
+            BorrowedItem { key1: "d", key2: Cow::Borrowed(b"b4"), key3: Path::new("path4") },
         };
 
         let diff = before.diff(&after);
@@ -153,24 +153,66 @@ fn test_extend() {
     assert_eq!(map.get(&TestKey1::new(&2)).unwrap().value, "w");
 }
 
+#[derive(Debug, Clone, Copy, PartialEq, Eq)]
+enum CompactnessChange {
+    /// The operation makes the map non-compact.
+    NoLongerCompact,
+    /// The operation makes the map compact.
+    BecomesCompact,
+    /// The operation doesn't change compactness.
+    NoChange,
+}
+
+impl CompactnessChange {
+    /// Applies this compactness change to the given compactness state.
+    fn apply(self, compactness: ValidateCompact) -> ValidateCompact {
+        match (compactness, self) {
+            (ValidateCompact::Compact, CompactnessChange::NoLongerCompact) => {
+                ValidateCompact::NonCompact
+            }
+            (
+                ValidateCompact::NonCompact,
+                CompactnessChange::BecomesCompact,
+            ) => ValidateCompact::Compact,
+            _ => compactness,
+        }
+    }
+}
+
 #[derive(Debug, Arbitrary)]
 enum Operation {
     // Make inserts a bit more common to try and fill up the map.
-    #[weight(3)]
+    #[weight(4)]
     InsertUnique(TestItem),
-    #[weight(2)]
+    #[weight(3)]
     InsertOverwrite(TestItem),
+    #[weight(2)]
     Get(u8),
+    #[weight(2)]
     Remove(u8),
+    #[weight(2)]
+    RetainValueContains(char, bool),
+    #[weight(2)]
+    RetainModulo(#[strategy(0..3_u8)] u8, #[strategy(1..4_u8)] u8, bool),
+    Clear,
 }
 
 impl Operation {
-    fn remains_compact(&self) -> bool {
+    fn compactness_change(&self) -> CompactnessChange {
         match self {
-            Operation::InsertUnique(_) | Operation::Get(_) => true,
+            Operation::InsertUnique(_) | Operation::Get(_) => {
+                CompactnessChange::NoChange
+            }
             // The act of removing items, including calls to insert_overwrite,
             // can make the map non-compact.
-            Operation::InsertOverwrite(_) | Operation::Remove(_) => false,
+            Operation::InsertOverwrite(_)
+            | Operation::Remove(_)
+            | Operation::RetainValueContains(_, _)
+            | Operation::RetainModulo(_, _, _) => {
+                CompactnessChange::NoLongerCompact
+            }
+            // Clear always makes the map compact (empty).
+            Operation::Clear => CompactnessChange::BecomesCompact,
         }
     }
 }
@@ -188,9 +230,7 @@ fn proptest_ops(
 
     // Now perform the operations on both maps.
     for op in ops {
-        if compactness == ValidateCompact::Compact && !op.remains_compact() {
-            compactness = ValidateCompact::NonCompact;
-        }
+        compactness = op.compactness_change().apply(compactness);
 
         match op {
             Operation::InsertUnique(item) => {
@@ -232,6 +272,35 @@ fn proptest_ops(
                 assert_eq!(map_res, naive_res);
                 map.validate(compactness).expect("map should be valid");
             }
+            Operation::RetainValueContains(ch, equals) => {
+                map.retain(|item| {
+                    let contains = item.value.contains(ch);
+                    if equals { contains } else { !contains }
+                });
+                naive_map.retain(|item| {
+                    let contains = item.value.contains(ch);
+                    if equals { contains } else { !contains }
+                });
+                map.validate(compactness).expect("map should be valid");
+            }
+            Operation::RetainModulo(a, b, equals) => {
+                let modulo = a + b;
+                let remainder = a;
+                map.retain(|item| {
+                    let matches = item.key1 % modulo == remainder;
+                    if equals { matches } else { !matches }
+                });
+                naive_map.retain(|item| {
+                    let matches = item.key1 % modulo == remainder;
+                    if equals { matches } else { !matches }
+                });
+                map.validate(compactness).expect("map should be valid");
+            }
+            Operation::Clear => {
+                map.clear();
+                naive_map.clear();
+                map.validate(compactness).expect("map should be valid");
+            }
         }
 
         // Check that the iterators work correctly.
@@ -424,10 +493,16 @@ fn insert_entry_panics_for_non_matching_key() {
 #[test]
 fn borrowed_item() {
     let mut map = IdHashMap::<BorrowedItem, HashBuilder, Alloc>::default();
-    let item1 =
-        BorrowedItem { key1: "foo", key2: b"foo", key3: Path::new("foo") };
-    let item2 =
-        BorrowedItem { key1: "bar", key2: b"bar", key3: Path::new("bar") };
+    let item1 = BorrowedItem {
+        key1: "foo",
+        key2: Cow::Borrowed(b"foo"),
+        key3: Path::new("foo"),
+    };
+    let item2 = BorrowedItem {
+        key1: "bar",
+        key2: Cow::Borrowed(b"bar"),
+        key3: Path::new("bar"),
+    };
 
     // Insert items.
     map.insert_unique(item1.clone()).unwrap();
@@ -450,12 +525,164 @@ fn borrowed_item() {
         format!("{map:?}")
     }
 
+    #[cfg(feature = "serde")]
+    fn serialize_as_map(
+        map: &IdHashMap<BorrowedItem<'_>, HashBuilder, Alloc>,
+    ) -> Result<String, iddqd_test_utils::serde_json::Error> {
+        let mut out: Vec<u8> = Vec::new();
+        let mut ser = iddqd_test_utils::serde_json::Serializer::new(&mut out);
+        id_hash_map::IdHashMapAsMap::serialize(map, &mut ser)?;
+        Ok(String::from_utf8(out)
+            .expect("serde_json should always emit valid UTF-8"))
+    }
+
     static DEBUG_OUTPUT: &str = "{\"foo\": BorrowedItem { \
         key1: \"foo\", key2: [102, 111, 111], key3: \"foo\" }, \
         \"bar\": BorrowedItem { \
         key1: \"bar\", key2: [98, 97, 114], key3: \"bar\" }}";
     assert_eq!(format!("{map:?}"), DEBUG_OUTPUT);
     assert_eq!(fmt_debug(&map), DEBUG_OUTPUT);
+
+    #[cfg(feature = "serde")]
+    {
+        let map_string = serialize_as_map(&map).unwrap();
+        let deserialized: IdHashMap<BorrowedItem<'_>, HashBuilder, Alloc> =
+            iddqd_test_utils::serde_json::from_str(&map_string).unwrap();
+        assert_eq!(map, deserialized);
+    }
+}
+
+#[test]
+fn test_retain_all() {
+    let mut map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(1, 'a', "x", "foo")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "bar")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "baz")).unwrap();
+
+    let original_len = map.len();
+    map.retain(|_| true);
+
+    assert_eq!(map.len(), original_len);
+    assert_eq!(map.len(), 3);
+    map.get(&TestKey1::new(&1)).expect("key 1 should be present");
+    map.get(&TestKey1::new(&2)).expect("key 2 should be present");
+    map.get(&TestKey1::new(&3)).expect("key 3 should be present");
+}
+
+#[test]
+fn test_retain_none() {
+    let mut map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(1, 'a', "x", "foo")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "bar")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "baz")).unwrap();
+
+    map.retain(|_| false);
+
+    assert_eq!(map.len(), 0);
+    assert!(map.is_empty());
+}
+
+#[test]
+fn test_retain_value_contains() {
+    let mut map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(1, 'a', "x", "foo")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "bar")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "baz")).unwrap();
+    map.insert_unique(TestItem::new(4, 'd', "w", "qux")).unwrap();
+
+    map.retain(|item| item.value.contains('a'));
+
+    assert_eq!(map.len(), 2);
+    map.get(&TestKey1::new(&2)).expect("key 2 (bar) should be present");
+    map.get(&TestKey1::new(&3)).expect("key 3 (baz) should be present");
+    assert!(
+        map.get(&TestKey1::new(&1)).is_none(),
+        "key 1 (foo) should be removed"
+    );
+    assert!(
+        map.get(&TestKey1::new(&4)).is_none(),
+        "key 4 (qux) should be removed"
+    );
+}
+
+#[test]
+fn test_retain_modulo() {
+    let mut map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.insert_unique(TestItem::new(0, 'a', "x", "v0")).unwrap();
+    map.insert_unique(TestItem::new(1, 'b', "y", "v1")).unwrap();
+    map.insert_unique(TestItem::new(2, 'c', "z", "v2")).unwrap();
+    map.insert_unique(TestItem::new(3, 'd', "w", "v3")).unwrap();
+    map.insert_unique(TestItem::new(4, 'e', "u", "v4")).unwrap();
+    map.insert_unique(TestItem::new(5, 'f', "t", "v5")).unwrap();
+
+    map.retain(|item| item.key1 % 3 == 1);
+
+    assert_eq!(map.len(), 2);
+    map.get(&TestKey1::new(&1)).expect("key 1 should be present");
+    map.get(&TestKey1::new(&4)).expect("key 4 should be present");
+    assert!(map.get(&TestKey1::new(&0)).is_none(), "key 0 should be removed");
+    assert!(map.get(&TestKey1::new(&2)).is_none(), "key 2 should be removed");
+    assert!(map.get(&TestKey1::new(&3)).is_none(), "key 3 should be removed");
+    assert!(map.get(&TestKey1::new(&5)).is_none(), "key 5 should be removed");
+
+    // Test with a larger map for miri coverage.
+    let mut large_map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    for i in 0..32_u8 {
+        large_map.insert_unique(TestItem::new(i, 'x', "y", "z")).unwrap();
+    }
+
+    large_map.retain(|item| item.key1 % 7 == 3);
+
+    for i in 0..32_u8 {
+        if i % 7 == 3 {
+            large_map
+                .get(&TestKey1::new(&i))
+                .unwrap_or_else(|| panic!("key {} should be present", i));
+        } else {
+            assert!(
+                large_map.get(&TestKey1::new(&i)).is_none(),
+                "key {} should be removed",
+                i
+            );
+        }
+    }
+}
+
+#[test]
+fn test_retain_empty_map() {
+    let mut map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.retain(|_| true);
+    assert!(map.is_empty());
+}
+
+#[test]
+fn test_clear_empty_map() {
+    let mut map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+    map.clear();
+    assert!(map.is_empty());
+    map.validate(ValidateCompact::Compact)
+        .expect("empty cleared map should be compact");
+}
+
+#[test]
+fn test_clear_makes_compact() {
+    let mut map = IdHashMap::<TestItem, HashBuilder, Alloc>::make_new();
+
+    // Add items
+    map.insert_unique(TestItem::new(1, 'a', "x", "v1")).unwrap();
+    map.insert_unique(TestItem::new(2, 'b', "y", "v2")).unwrap();
+    map.insert_unique(TestItem::new(3, 'c', "z", "v3")).unwrap();
+
+    // Remove an item to make it non-compact
+    map.remove(&TestKey1::new(&2));
+    map.validate(ValidateCompact::NonCompact)
+        .expect("map should be valid but non-compact");
+
+    // Clear should make it compact again
+    map.clear();
+    assert!(map.is_empty());
+    map.validate(ValidateCompact::Compact)
+        .expect("cleared map should be compact");
 }
 
 mod macro_tests {
